# Anurag Bishnoi, anurag.2357@gmail.com

# Run this code to generate all 350 distance-2 ovoids in H(4, 1) taken as a subgeometry 
#of H(4)^D.

# These distance-2 ovoids are partial distance-2 ovoids in H(4)^D which we will try to
# extend. 

def read_lines_from_file(name):
    """
    Read a file containing the lines of the geometry.

    Args: 
    name -- the file name (each line of the file should have integers separated by a space)

    Returns:
    a list whose each element is a line of the geometry stored as a list of points incident to the line
    """
    L = []
    f = open(name, 'r')
    for l in f.readlines():
        l = l.split()
        L.append([int(x) for x in l])
    f.close()
    return L

def read_group_from_file(name):
    """
    Read a file containing the automorphism group of the geometry.
  
    Args: 
    name -- the file name

    Returns: 
    a GAP object that is the automorphism group of the geometry 
    """
    f = open(name, 'r')
    s = f.read()
    s = s.replace("\\","") # the groups written in the file using GAP contain these characters
    f.close()
    return gap(s)


def ovoids(P,L):
    """
    Find all 1-ovoids in the geometry. 

    Args:
    P -- the vertices (points) of the hypergraph (geometry) 
    L -- the edges (lines) of the hypergraph (geometry)

    Returns:
    an iterator to the exact hitting sets (1-ovoids) in the hypergraph (geometry).
    """
    map = dict()
    for p in P:
        map[p] = []
    for i in range(len(L)):
        for p in L[i]:
            map[p].append(i)
    E = [map[p] for p in P]
    for match in DLXCPP(E):
        yield [P[i] for i in match]



# the next two functions are defined to compute a subgeometry of H(4)^D isomorphic to H(4,1).

# A set X of points is a subspace if for every two collinear points a, b in X all the 
# points of the line ab are contained in X. 

# Given a set S of points, the span of S is the smallest subspace that contains S. 

# A set X of points is convex if for every pair of points a, b in X all the points 
# lying on all shortest paths between a and b are also contained in X. 

# Given a set S of points, the convex closure of S of depth j is the smallest 
# subset X of points containing S such that for every a, b in X with d(a, b) <= j
# every point on every shortest path connecting a and b is contained in X. 

def subspace(S, points, lines):
    """
    Find the smallest subspace containing the set S of points. 

    Args:
    S -- a subset of points  
    points -- the points of the geometry
    lines -- the lines of the geometry

    Returns:
    the subset span(S) of points 
    """

    status = true
    n = len(points)
    t = len(lines[0]) # the number of points on a line
    while status and len(S) <= n:
        status = false
        for l in lines:
            k = len([x for x in l if x in S])
            if k >= 2 and k < t:
                S = S.union(set(l))
                status = true
    return S


def convex(S, depth, points, lines, T, D):
    """
    Find the convex closure of S of given depth.

    Args:
    S -- a subset of points  
    depth -- the maximum distance considered for convex closure
    points -- the points of the geometry
    lines -- the lines of the geometry
    T -- collinearity graph of the geometry
    D -- dictionary containing distance between all pairs of points 

    Returns:
    the convex closure of given depth
    """

    status = true
    n = len(points)
    S = subspace(S, points, lines)
    while status and len(S) < n:
        status = false
        for x in S:
            for y in S:
                if x != y and D[x][y] <= depth:
                    for z in T.neighbors(x):
                        if D[z][y] == D[x][y] - 1 and not z in S:
                            S = S.union([z]) 
                            status = true
        S = subspace(S, points, lines)
    return S
        
                   

lines = read_lines_from_file("lines")

G = read_group_from_file("group")
edges = []
for l in lines:
    edges.extend(Combinations(l,2))
T = Graph(edges)
points = T.vertices()
D = T.distance_all_pairs()

# We now construct a subgeometry of H(4)^D isomorphic to H(4, 1). 
p = 1
a = [x for x in points if D[x][p] == 3][0]
b = [x for x in T.neighbors(a) if D[x][p] == 2][0]
c = [x for x in T.neighbors(a) if D[x][p] == 2][1]

P = list(convex({p,a,b,c}, 2, points, lines, T, D)) # points of H(4,1)
L = [l for l in lines if all([x in P for x in l])] # lines of H(4,1)
H = gap.Stabilizer(G, sorted(P), 'OnSets') # isomorphic to the full automorphism group of H(4,1)

# write some data to be used later
f = open("lines_for_gap.g", 'w')
f.write("points := [1..1365];" + "\n" + "lines := " + str(lines) + ";" + "\n" + "H := " + str(H) + ";")
f.close()


gap('LoadPackage("grape")')
Ov = dict() # to store the partial distance-2 ovoids
n = 0
bal = 18534400 # number of perfect matchings in the incidence graph of PG(2, 4)
O = ovoids(P,L) # the iterator to ovoids in H(4, 1)

print "Generating ovoids ..."
while bal > 0:
    x = sorted(O.next())
    s = 241920/gap.Size(gap.Stabiliser(H, x, 'OnSets'))
    if s > bal: 
        continue
    x = gap.SmallestImageSet(H, x) 
    if Ov.get(x) == None: 
        Ov[x] = s
        bal = bal - s
        n += 1 
        print n, bal

partials = Ov.keys() 
f = open("all_ovoids_of_H_4_1_gap.g", 'w')
f.write("ovoids := " + str(partials) + ";")
f.close()




